#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h>

#include <kernel/tty.h>

#include "vga.h"
#include "../../kernel/asm_helper.h" // FIXME

static const size_t VGA_WIDTH = 80;
static const size_t VGA_HEIGHT = 25;
static uint16_t* const VGA_MEMORY = (uint16_t*) 0xB8000;

static size_t terminal_row;
static size_t terminal_column;
static uint8_t terminal_color;
static uint16_t* terminal_buffer;

void terminal_initialize(void) {
	terminal_row = 0;
	terminal_column = 0;
	terminal_color = vga_entry_color(VGA_COLOR_LIGHT_GREY, VGA_COLOR_BLACK);
	terminal_buffer = VGA_MEMORY;
	for (size_t y = 0; y < VGA_HEIGHT; y++) {
		for (size_t x = 0; x < VGA_WIDTH; x++) {
			const size_t index = y * VGA_WIDTH + x;
			terminal_buffer[index] = vga_entry(' ', terminal_color);
		}
	}
}

void terminal_setcolor(uint8_t color) {
	terminal_color = color;
}

void terminal_putentryat(unsigned char c, uint8_t color, size_t x, size_t y) {
	const size_t index = y * VGA_WIDTH + x;
	terminal_buffer[index] = vga_entry(c, color);
}

// Updates the hardware cursor.
static void move_cursor()
{
   // The screen is 80 characters wide...
   uint16_t cursorLocation = terminal_row * 80 + terminal_column;
   outb(0x3D4, 14);                  // Tell the VGA board we are setting the high cursor byte.
   outb(0x3D5, cursorLocation >> 8); // Send the high cursor byte.
   outb(0x3D4, 15);                  // Tell the VGA board we are setting the low cursor byte.
   outb(0x3D5, cursorLocation);      // Send the low cursor byte.
}

void terminal_scroll(int line) {
	// Get a space character with the default colour attributes.
	int attributeByte = (0 /*black*/ << 4) | (15 /*white*/ & 0x0F);
	uint16_t blank = 0x20 /* space */ | (attributeByte << 8);

        // Row 25 is the end, this means we need to scroll up
        if(terminal_row >= 25)
        {
            // Move the current text chunk that makes up the screen
            // back in the buffer by a line
            int i;
            for (i = 0*80; i < 24*80; i++)
            {
                terminal_buffer[i] = terminal_buffer[i+80];
            }
     
            // The last line should now be blank. Do this by writing
            // 80 spaces to it.
            for (i = 24*80; i < 25*80; i++)
            {
                terminal_buffer[i] = blank;
            }
            // The cursor should now be on the last line.
            terminal_row = 24;
        }
}

 
void terminal_delete_last_line() {
	int x, *ptr;
 
	for(x = 0; x < VGA_WIDTH * 2; x++) {
		ptr = 0xB8000 + (VGA_WIDTH * 2) * (VGA_HEIGHT - 1) + x;
		*ptr = 0;
	}
}

void terminal_putchar(char c) {
	unsigned char uc = c;
	terminal_putentryat(uc, terminal_color, terminal_column, terminal_row);
	if (c != '\n') {
	  if (++terminal_column == VGA_WIDTH) {
	    terminal_column = 0;
	    if (++terminal_row == VGA_HEIGHT) {
             	for(int line = 1; line <= VGA_HEIGHT - 1; line++) {
			terminal_scroll(line);
		}
        	terminal_delete_last_line();
		terminal_row = VGA_HEIGHT - 1;
	    }
	  }
	} else {
	  terminal_row += 1;
	  terminal_column = 0;
	}
	terminal_scroll(terminal_row);
	move_cursor();
}

void terminal_write(const char* data, size_t size) {
	for (size_t i = 0; i < size; i++)
		terminal_putchar(data[i]);
}

void terminal_writestring(const char* data) {
	terminal_write(data, strlen(data));
}
